#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <byteswap.h>
#include <string.h>
#include <arpa/inet.h>
#include <ulogd/ulogd.h>
#include <ulogd/conffile.h>
#include <db.h>

#define ENTRY_VERSION	1

#define MAX_STRING	80

struct key {
	unsigned int	sec;
	unsigned int	usec;
};

struct entry {
	unsigned int	version;
	unsigned int	oob_time_sec;
	unsigned int	oob_time_usec;
	unsigned int	ip_saddr;
	unsigned int	ip_daddr;
	unsigned char	ip_protocol;
	unsigned char	tcp_flags;
	unsigned short	ip_totlen;
	unsigned short	tcp_sport;
	unsigned short	tcp_dport;
	unsigned short	udp_sport;
	unsigned short	udp_dport;
	unsigned short	udp_len;
	unsigned char	icmp_type;
	unsigned char	icmp_code;
	unsigned short	icmp_echoid;
	unsigned short	icmp_echoseq;
	unsigned int	icmp_gateway;
	unsigned short	icmp_fragmtu;
	char		raw_mac[MAX_STRING];
};

#define OFFSET(x)	offsetof(struct entry, x)

static const struct field {
	char		*name;
	unsigned int	offset;
	unsigned int	bit;
} fields[] = {
	{ "oob.time.sec",	OFFSET(oob_time_sec), },
	{ "oob.time.usec",	OFFSET(oob_time_usec), },
	{ "ip.saddr",		OFFSET(ip_saddr), },
	{ "ip.daddr",		OFFSET(ip_daddr), },
	{ "ip.protocol",	OFFSET(ip_protocol), },
	{ "ip.totlen",		OFFSET(ip_totlen), },
	{ "tcp.sport",		OFFSET(tcp_sport), },
	{ "tcp.dport",		OFFSET(tcp_dport), },
	{ "tcp.fin",		OFFSET(tcp_flags), 0 },
	{ "tcp.syn",		OFFSET(tcp_flags), 1 },
	{ "tcp.rst",		OFFSET(tcp_flags), 2 },
	{ "tcp.psh",		OFFSET(tcp_flags), 3 },
	{ "tcp.ack",		OFFSET(tcp_flags), 4 },
	{ "tcp.urg",		OFFSET(tcp_flags), 5 },
	{ "udp.sport",		OFFSET(udp_sport), },
	{ "udp.dport",		OFFSET(udp_dport), },
	{ "udp.len",		OFFSET(udp_len), },
	{ "icmp.type",		OFFSET(icmp_type), },
	{ "icmp.code",		OFFSET(icmp_code), },
	{ "icmp.echoid",	OFFSET(icmp_echoid), },
	{ "icmp.echoseq",	OFFSET(icmp_echoseq), },
	{ "icmp.gateway",	OFFSET(icmp_gateway), },
	{ "icmp.fragmtu",	OFFSET(icmp_fragmtu), },
	{ "raw.mac",		OFFSET(raw_mac), },
	{ NULL, 0, 0 }
};

#undef OFFSET

static DB_ENV *env;
static DB *db;
static int buffer;

static config_entry_t db_ce = { 
	.key = "db", 
	.type = CONFIG_TYPE_STRING,
	.options = CONFIG_OPT_MANDATORY,
};

static config_entry_t env_ce = { 
	.next = &db_ce,
	.key = "env", 
	.type = CONFIG_TYPE_STRING,
	.options = CONFIG_OPT_MANDATORY,
};

static config_entry_t buffer_ce = { 
	.next = &env_ce,
	.key = "buffer",
	.type = CONFIG_TYPE_INT,
	.options = CONFIG_OPT_MANDATORY,
};

static int _bdb_output(ulog_iret_t *result)
{
	DBT key, data;
	struct key k;
	struct entry e = { ENTRY_VERSION, };
	const struct field *f;

	for(f = fields; f->name; f++) {
		ulog_iret_t *res;
		int id;

		id = keyh_getid(f->name);
		if (!id) {
			ulogd_log(ULOGD_NOTICE, "unknown key %s\n", f->name);
			continue;
		}

		res = keyh_getres(id);
		if (!res || !IS_VALID((*res)))
			continue;

		switch(res->type) {
			case ULOGD_RET_INT8:
				*(char *)((void *)&e + f->offset) = res->value.i8;
				break;
			case ULOGD_RET_INT16:
				*(short *)((void *)&e + f->offset) = res->value.i16;
				break;
			case ULOGD_RET_INT32:
				*(int *)((void *)&e + f->offset) = res->value.i32;
				break;
			case ULOGD_RET_INT64:
				*(int64_t *)((void *)&e + f->offset) = res->value.i64;
				break;
			case ULOGD_RET_UINT8:
				*(unsigned char *)((void *)&e + f->offset) = res->value.ui8;
				break;
			case ULOGD_RET_UINT16:
				*(unsigned short *)((void *)&e + f->offset) = res->value.ui16;
				break;
			case ULOGD_RET_IPADDR:
			case ULOGD_RET_UINT32:
				*(unsigned int *)((void *)&e + f->offset) = res->value.ui32;
				break;
			case ULOGD_RET_UINT64:
				*(uint64_t *)((void *)&e + f->offset) = res->value.ui64;
				break;
			case ULOGD_RET_BOOL:
				*(char *)((void *)&e + f->offset) |=
					(res->value.b ? 1 : 0) << f->bit;
				break;
			case ULOGD_RET_STRING:
				strncpy((char *)((void *)&e + f->offset), res->value.ptr, MAX_STRING);
				break;
			default:
				ulogd_log(ULOGD_NOTICE,
				          "unknown type %d for %s\n",
				          res->type, res->key);
				break;
		}
	}

	k.sec = bswap_32(e.oob_time_sec);
	k.usec = bswap_32(e.oob_time_usec);

	memset(&key, 0, sizeof(DBT));
	key.data = &k;
	key.size = sizeof k;

	memset(&data, 0, sizeof(DBT));
	data.data = &e;
	data.size = sizeof e;

	db->put(db, NULL, &key, &data, 0);
	if (++buffer >= buffer_ce.u.value) {
		buffer = 0;
		db->sync(db, 0);
	}

	return 0;
}

static int _bdb_init(void)
{
	int ret;

	config_parse_file("BDB", &buffer_ce);

	buffer = 0;

	ret = db_env_create(&env, 0);
	if (ret) {
		ulogd_log(ULOGD_ERROR, "1 %s\n", db_strerror(ret));
		return 1;
	}

	ret = env->open(env, env_ce.u.string,
	                DB_CREATE | DB_INIT_LOCK | DB_INIT_MPOOL, 0);
	if (ret) {
		ulogd_log(ULOGD_ERROR, "2 %s\n", db_strerror(ret));
		env->close(env, 0);
		return 1;
	}

	ret = db_create(&db, env, 0);
	if (ret) {
		ulogd_log(ULOGD_ERROR, "3 %s\n", db_strerror(ret));
		env->close(env, 0);
		return 1;
	}

	db->set_flags(db, DB_DUP);
	if (ret) {
		ulogd_log(ULOGD_ERROR, "4 %s\n", db_strerror(ret));
		env->close(env, 0);
		return 1;
	}

	ret = db->open(db, NULL, db_ce.u.string, NULL, DB_BTREE,
	                DB_CREATE, 0);
	if (ret) {
		ulogd_log(ULOGD_ERROR, "5 %s\n", db_strerror(ret));
		db->close(db, 0);
		env->close(env, 0);
		return 1;
	}

	return 0;
}

static void _bdb_fini(void)
{
	db->close(db, 0);
	env->close(env, 0);
}

static ulog_output_t _bdb_plugin = { 
	.name = "bdb", 
	.output = &_bdb_output, 
	.init = &_bdb_init,
	.fini = &_bdb_fini,
};

void __attribute__ ((constructor)) _module_init(void);
void __attribute__ ((constructor)) _module_init(void)
{
	register_output(&_bdb_plugin);
}
